
;ROUND equ 8
;SIZE_X = 512
;SIZE_Y = 512

CATMULL_SHIFT equ 8

;------------------------------------------------------------------------
;- Procedure drawing textured triangle using Catmull Z-buffer algorithm -
;------------------------------------------------------------------------
tex_triangle_z:
;----------in - eax - x1 shl 16 + y1
;-------------- ebx - x2 shl 16 + y2
;---------------ecx - x3 shl 16 + y3
;---------------edx - pointer to 32 bit integer Z-buffer
;---------------esi - pointer to 32 bit texture buffer
;---------------edi - pointer to 32 bit screen buffer
;-------------stack - texture coordinates
;------------------ - z coordinates
.tex_x1 equ  ebp+4
.tex_y1 equ  ebp+6
.tex_x2 equ  ebp+8
.tex_y2 equ  ebp+10
.tex_x3 equ  ebp+12
.tex_y3 equ  ebp+14
.z1	equ  word[ebp+16]
.z2	equ  word[ebp+18]
.z3	equ  word[ebp+20]

.tex_ptr equ dword[ebp-4]	 ; pointer to texture
.z_ptr	 equ dword[ebp-8]	 ; pointer to z-buffer
.x1	 equ word[ebp-10]
.y1	 equ word[ebp-12]
.x2	 equ word[ebp-14]
.y2	 equ word[ebp-16]
.x3	 equ word[ebp-18]
.y3	 equ word[ebp-20]

.dx12	  equ dword[ebp-24]
.tex_dx12 equ dword[ebp-28]
.tex_dy12 equ dword[ebp-32]
.dz12	  equ dword[ebp-36]

.dx13	  equ dword[ebp-40]
.tex_dx13 equ dword[ebp-44]
.tex_dy13 equ dword[ebp-48]
.dz13	  equ dword[ebp-52]

.dx23	  equ dword[ebp-56]
.tex_dx23 equ dword[ebp-60]
.tex_dy23 equ dword[ebp-64]
.dz23	  equ dword[ebp-68]

.scan_x1  equ dword[ebp-72]
.scan_x2  equ dword[ebp-76]
.scan_y1  equ dword[ebp-80]
.scan_y2  equ dword[ebp-84]
.cz1	  equ dword[ebp-88]
.cz2	  equ dword[ebp-92]

	mov	ebp,esp

	push	esi		  ; store memory pointers
	push	edx
.tt_sort3:
	cmp	ax,bx			  ;sort all parameters
	jle	.tt_sort1
	xchg	eax,ebx
	mov	edx,dword [.tex_x1]
	xchg	edx,dword [.tex_x2]
	mov	dword[.tex_x1],edx
	mov	dx,.z1
	xchg	dx,.z2
	mov	.z1,dx
.tt_sort1:
	cmp	bx,cx
	jle	.tt_sort2
	xchg	ebx,ecx
	mov	edx,dword [.tex_x2]
	xchg	edx,dword [.tex_x3]
	mov	dword [.tex_x2],edx
	mov	dx,.z2
	xchg	dx,.z3
	mov	.z2,dx
	jmp	.tt_sort3
.tt_sort2:

	push	eax	; and store to user friendly variables
	push	ebx
	push	ecx

	mov	 edx,80008000h	; eax,ebx,ecx are ANDd together into edx which means that
	and	 edx,ebx	; if *all* of them are negative a sign flag is raised
	and	 edx,ecx
	and	 edx,eax
	test	 edx,80008000h	; Check both X&Y at once
	jne	 .tt_loop2_end
	cmp	 ax,SIZE_Y
	jl	 @f
	cmp	 bx,SIZE_Y
	jl	 @f
	cmp	 cx,SIZE_Y
	jl	 @f
	ror	 eax,16
	ror	 ebx,16
	ror	 ecx,16
	cmp	 ax,SIZE_X
	jl	 @f
	cmp	 bx,SIZE_X
	jl	 @f
	cmp	 cx,SIZE_X
	jl	 @f
	jmp	 .tt_loop2_end
     @@:
	mov	 eax,dword[.tex_x1] ; texture coords must be in [0..TEX_X(Y)]
	mov	 ebx,dword[.tex_x2]
	mov	 ecx,dword[.tex_x3]
	mov	 edx,eax
	or	 edx,ebx
	or	 edx,ecx
	test	 edx,80008000h
	jne	 .tt_loop2_end
	cmp	 ax,TEX_X
	jge	 .tt_loop2_end
	cmp	 bx,TEX_X
	jge	 .tt_loop2_end
	cmp	 cx,TEX_X
	jge	 .tt_loop2_end
	ror	 eax,16
	ror	 ebx,16
	ror	 ecx,16
	cmp	 ax,TEX_Y
	jge	 .tt_loop2_end
	cmp	 bx,TEX_Y
	jge	 .tt_loop2_end
	cmp	 cx,TEX_Y
	jge	 .tt_loop2_end


	movsx	 ebx,.y2    ; calc delta
	sub	 bx,.y1
	jnz	 .tt_dx12_make
	xor	 edx,edx
	mov	 ecx,4
    @@:
	push	 edx
	loop	 @b
	jmp	 .tt_dx12_done
    .tt_dx12_make:
	mov	 ax,.x2
	sub	 ax,.x1
	cwde
	shl	 eax,ROUND
	cdq
	idiv	 ebx
       ; mov      .dx12,eax                      ; dx12 = (x2-x1)/(y2-y1)
	push	 eax

	mov	 ax,word[.tex_x2]
	sub	 ax,word[.tex_x1]
	cwde
	shl	 eax,ROUND
	cdq
	idiv	 ebx
       ; mov     [.tex_dx12],eax               ; tex_dx12 = (tex_x2-tex_x1)/(y2-y1)
	push	 eax

	mov	 ax,word[.tex_y2]
	sub	 ax,word[.tex_y1]
	cwde
	shl	 eax,ROUND
	cdq
	idiv	 ebx
     ;   mov    [.tex_dy12],eax              ; tex_dy12 = (tex_y2-tex_y1)/(y2-y1)
	push	 eax

	mov	 ax,.z2
	sub	 ax,.z1
	cwde
	shl	 eax,CATMULL_SHIFT
	cdq
	idiv	 ebx
	push	 eax
   .tt_dx12_done:

	movsx	 ebx,.y3    ; calc delta
	sub	 bx,.y1
	jnz	 .tt_dx13_make
	xor	 edx,edx
	mov	 ecx,4
    @@:
	push	 edx
	loop	 @b
	jmp	 .tt_dx13_done
    .tt_dx13_make:
	mov	 ax,.x3
	sub	 ax,.x1
	cwde
	shl	 eax,ROUND
	cdq
	idiv	 ebx
       ; mov      .dx12,eax                      ; dx13 = (x3-x1)/(y3-y1)
	push	  eax

	mov	 ax,word[.tex_x3]
	sub	 ax,word[.tex_x1]
	cwde
	shl	 eax,ROUND
	cdq
	idiv	 ebx
       ; mov     [.tex_dx12],eax               ; tex_dx13 = (tex_x3-tex_x1)/(y3-y1)
	push	 eax

	mov	 ax,word[.tex_y3]
	sub	 ax,word[.tex_y1]
	cwde
	shl	 eax,ROUND
	cdq
	idiv	 ebx
     ;   mov    [.tex_dy12],eax              ; tex_dy13 = (tex_y3-tex_y1)/(y3-y1)
	push	 eax

	mov	 ax,.z3
	sub	 ax,.z1
	cwde
	shl	 eax,CATMULL_SHIFT
	cdq
	idiv	 ebx
	push	 eax
   .tt_dx13_done:

	mov	 bx,.y3    ; calc delta
	sub	 bx,.y2
	jnz	 .tt_dx23_make
	xor	 edx,edx
	mov	 ecx,4
    @@:
	push	 edx
	loop	 @b
	jmp	 .tt_dx23_done
    .tt_dx23_make:
	mov	 ax,.x3
	sub	 ax,.x2
	cwde
	shl	 eax,ROUND
	cdq
	movzx	 ebx,bx
	idiv	 ebx
       ; mov      .dx23,eax                      ; dx23 = (x3-x2)/(y3-y2)
	push	  eax

	mov	 ax,word[.tex_x3]
	sub	 ax,word[.tex_x2]
	cwde
	shl	 eax,ROUND
	cdq
	idiv	 ebx
       ; mov     [.tex_dx23],eax               ; tex_dx23 = (tex_x3-tex_x2)/(y3-y2)
	push	 eax

	mov	 ax,word[.tex_y3]
	sub	 ax,word[.tex_y2]
	cwde
	shl	 eax,ROUND
	cdq
	idiv	 ebx
     ;   mov    [.tex_dy23],eax              ; tex_dy23 = (tex_y3-tex_y2)/(y3-y2)
	push	 eax

	mov	 ax,.z3
	sub	 ax,.z2
	cwde
	shl	 eax,CATMULL_SHIFT
	cdq
	idiv	 ebx
	push	 eax
   .tt_dx23_done:

	movsx	 eax,.x1     ;eax - cur x1
	shl	 eax,ROUND   ;ebx - cur x2
	mov	 ebx,eax

	movsx	 edx, word[.tex_x1]
	shl	 edx,ROUND
      ;  mov [.scan_x1],edx
      ;  mov [.scan_x2],edx
	push	 edx
	push	 edx
	movsx	 edx, word[.tex_y1]
	shl	 edx,ROUND
       ; mov [.scan_y1],edx
       ; mov [.scan_y2],edx
	push	 edx
	push	 edx
	movsx	 edx,.z1
	shl	 edx,CATMULL_SHIFT
	push	 edx
	push	 edx
	mov	 cx,.y1
	cmp	 cx,.y2
	jge	 .tt_loop1_end

   .tt_loop1:
	pushad

	push	.z_ptr
	push	.cz1	  ; z coords shifted shl catmull_shift
	push	.cz2
	push	.scan_y2
	push	.scan_x2
	push	.scan_y1
	push	.scan_x1
	push	esi		  ;[.tex_ptr]

	push	cx
	sar	ebx,ROUND
	push	bx
	sar	eax,ROUND
	push	ax
	call	textured_line_z

	popad
	mov	edx,.dz13
	add	.cz1,edx
	mov	edx,.dz12
	add	.cz2,edx

	mov	edx, .tex_dx13
	add	.scan_x1, edx
	mov	edx, .tex_dx12
	add	.scan_x2, edx
	mov	edx, .tex_dy13
	add	.scan_y1, edx
	mov	edx, .tex_dy12
	add	.scan_y2, edx

	add	eax, .dx13
	add	ebx, .dx12
	inc	cx
	cmp	cx,.y2
	jl	.tt_loop1

  .tt_loop1_end:


	mov	cx,.y2
	cmp	cx,.y3
	jge	.tt_loop2_end

	movsx	ebx,.x2
	shl	ebx,ROUND
	movsx	edx,.z2
	shl	edx,CATMULL_SHIFT
	mov	.cz2,edx
	movzx	edx, word [.tex_x2]
	shl	edx,ROUND
	mov	.scan_x2,edx
	movzx	edx, word[.tex_y2]
	shl	edx,ROUND
	mov	.scan_y2,edx

.tt_loop2:

	pushad

	push	.z_ptr
	push	.cz1	  ; z coords shifted shl catmull_shift
	push	.cz2

	push	.scan_y2
	push	.scan_x2
	push	.scan_y1
	push	.scan_x1
	push	esi		  ;[.tex_ptr]

	push	cx
	sar	ebx,ROUND
	push	bx
	sar	eax,ROUND
	push	ax
	call	textured_line_z

	popad


	mov	edx,.dz13
	add	.cz1,edx
	mov	edx,.dz23
	add	.cz2,edx

	mov	edx, .tex_dx13
	add	.scan_x1, edx
	mov	edx, .tex_dx23
	add	.scan_x2, edx
	mov	edx, .tex_dy13
	add	.scan_y1, edx
	mov	edx, .tex_dy23
	add	.scan_y2, edx

	add	eax, .dx13
	add	ebx, .dx23
	inc	cx
	cmp	cx,.y3
	jl	.tt_loop2

.tt_loop2_end:

.tt_end:
	mov esp,ebp

ret 18

textured_line_z:
;-----in -edi screen buffer pointer
;------------ stack:
  .x1 equ word [ebp+4]
  .x2 equ word [ebp+6]
  .y  equ word [ebp+8]

  .tex_ptr equ dword [ebp+10]
  .tex_x1  equ ebp+14
  .tex_y1  equ ebp+18
  .tex_x2  equ ebp+22
  .tex_y2  equ ebp+26
  .z2	   equ dword [ebp+30]	  ;z1, z2  coords shifted shl CATMULL_SHIFT
  .z1	   equ dword [ebp+34]
  .z_ptr   equ dword [ebp+38]

  .tex_dy  equ dword [ebp-4]
  .tex_dx  equ dword [ebp-8]
  .dz	   equ dword [ebp-12]
  .cz	   equ dword [ebp-16]
  .c_tex_x equ dword [ebp-20]  ; current tex x
  .c_tex_y equ dword [ebp-24]
  .rest    equ ebp-28
  .u	   equ ebp-32
  .v	   equ ebp-36

  .tex_dyI    equ dword ebx	    ;      sse_repository
  .tex_dyII   equ dword ebx+4	    ;      sse_repository
  .tex_dyIII  equ dword ebx+8	    ;      sse_repository
  .tex_dyIV   equ dword ebx+12	    ;      sse_repository
  .tex_dxI    equ dword ebx+16	    ;      sse_repository
  .tex_dxII   equ dword ebx+20	    ;      sse_repository
  .tex_dxIII  equ dword ebx+24	    ;      sse_repository
  .tex_dxIV   equ dword ebx+28	    ;      sse_repository
  .dzI	      equ dword ebx+32	    ;      sse_repository
  .dzII       equ dword ebx+36	    ;      sse_repository
  .dzIII      equ dword ebx+40	    ;      sse_repository
  .dzIV       equ dword ebx+44	    ;      sse_repository
  .cyI	      equ dword ebx+48	    ;      sse_repository
  .cyII       equ dword ebx+52	    ;      sse_repository
  .cyIII      equ dword ebx+56	    ;      sse_repository
  .cyIV       equ dword ebx+60	    ;      sse_repository
  .cxI	      equ dword ebx+64	    ;      sse_repository
  .cxII       equ dword ebx+68	    ;      sse_repository
  .cxIII      equ dword ebx+72	    ;      sse_repository
  .cxIV       equ dword ebx+76	    ;      sse_repository
  .czI	      equ dword ebx+80	    ;      sse_repository
  .czII       equ dword ebx+84	    ;      sse_repository
  .czIII      equ dword ebx+88	    ;      sse_repository
  .czIV       equ dword ebx+92	    ;      sse_repository
  .tex_ptrI   equ dword ebx+96	    ;      sse_repository
  .tex_ptrII  equ dword ebx+100     ;      sse_repository
  .tex_ptrIII equ dword ebx+104     ;      sse_repository
  .tex_ptrIV  equ dword ebx+108     ;      sse_repository
  .colI       equ dword ebx+112     ;      sse_repository
  .colII      equ dword ebx+116     ;      sse_repository
  .colIII     equ dword ebx+120     ;      sse_repository
  .colIV      equ dword ebx+124     ;      sse_repository
  .colIb      equ dword ebx+128
  .colIIb     equ dword ebx+132
  .colIIIb    equ dword ebx+136
  .colIVb     equ dword ebx+140
  .colIc      equ dword ebx+144
  .colIIc     equ dword ebx+148
  .colIIIc    equ dword ebx+152
  .colIVc     equ dword ebx+156
  .colId      equ dword ebx+160
  .colIId     equ dword ebx+164
  .colIIId    equ dword ebx+168
  .colIVd     equ dword ebx+172
  .u1	      equ  ebx+176
  .u2	      equ  ebx+180
  .u3	      equ  ebx+184
  .u4	      equ  ebx+188
  .v1	      equ  ebx+192
  .v2	      equ  ebx+196
  .v3	      equ  ebx+200
  .v4	      equ  ebx+204


	mov	ebp,esp
	mov	ax,.y
	or	ax,ax
	jl	.tl_quit
	cmp	ax,SIZE_Y
	jge	.tl_quit

	mov	ax,.x1
	cmp	ax,.x2
	je	.tl_quit
	jl	.tl_ok

	xchg	ax,.x2	  ; sort params
	mov	.x1,ax

	movq	mm0,[.tex_x1]
	movq	mm1,[.tex_x2]
	movq	[.tex_x2],mm0
	movq	[.tex_x1],mm1


	mov	eax,.z1
	xchg	eax,.z2
	mov	.z1,eax

     .tl_ok:
	cmp	.x1,SIZE_X
	jge	.tl_quit
	cmp	.x2,0
	jle	.tl_quit

	mov	bx,.x2
	sub	bx,.x1
	movsx	ebx,bx

	mov	eax,dword[.tex_y2]	 ; calc .dty
	sub	eax,dword[.tex_y1]
	cdq
	idiv	ebx
	push	eax

	mov	eax,dword[.tex_x2]	 ; calc .dtx
	sub	eax,dword[.tex_x1]
	cdq
	idiv	ebx
	push	eax

	mov	eax,.z2       ; calc .dz
	sub	eax,.z1
	cdq
	idiv	ebx
	push	eax

	cmp    .x1,0	      ; clipping
	jg     @f

	movsx	ebx,.x1
	neg	ebx
	imul	ebx	      ; eax = .dz * abs(.x1)
	add	.z1,eax
	mov	.x1,0

	mov	eax,.tex_dy
	imul	ebx
	add	dword[.tex_y1],eax

	mov	eax,.tex_dx
	imul	ebx
	add	dword[.tex_x1],eax

      @@:
	cmp	.x2,SIZE_X
	jl	@f
	mov	.x2,SIZE_X
      @@:

	movsx	ebx,.y	      ; calc mem begin in buffers
	mov	eax,SIZE_X
	mul	ebx
	movsx	ebx,.x1
	add	eax,ebx

	shl	eax,2
	add	edi,eax 	      ; edi - scr buff
	add	.z_ptr,eax	      ; z buffer pointer

	mov	cx,.x2
	sub	cx,.x1
	jz	.tl_quit
	movzx	ecx,cx

	mov	eax,dword[.tex_x1]
   ;     mov     ebx,dword[.tex_y1]
	push	.z1	      ; .cz
	push	eax	      ; .c_tex_x
	push	dword[.tex_y1]
	sub	esp,100

	mov	 esi,.z_ptr

	mov	  word [.rest],0
	shr	  ecx,1
	jnc	  @f
	add	  word [.rest],1
    @@:
	shr	  ecx,1
	jnc	  @f
	add	  word	[.rest],2
    @@:
	cmp	  ecx,0
	jz	  .draw_last1

 ;       movd    xmm0,.dz
 ;       shufps  xmm0,xmm0

	 mov	  ebx,sse_repository
	 movd	  xmm1,eax
	 shufps   xmm1,xmm1,0
	 movd	  xmm0,.tex_dx
	 shufps   xmm0,xmm0,0
	 movaps   xmm7,xmm0		   ; xmm7 -> dx
	 cvtdq2ps xmm0,xmm0
	 mulps	  xmm0,[muler0123]
	 cvtps2dq xmm0,xmm0
	 paddd	  xmm1,xmm0
	 movaps   [.cxI],xmm1
	 pslld	  xmm7,2
	 movaps   [.tex_dxI],xmm7

	 movd	  xmm1,dword[.tex_y1]
	 shufps   xmm1,xmm1,0
	 movd	  xmm0,.tex_dy
	 shufps   xmm0,xmm0,0
	 movaps   xmm7,xmm0		   ; xmm7 -> dy
	 cvtdq2ps xmm0,xmm0
	 mulps	  xmm0,[muler0123]
	 cvtps2dq xmm0,xmm0
	 paddd	  xmm1,xmm0
	 movaps   [.cyI],xmm1
	 pslld	  xmm7,2
	 movaps   [.tex_dyI],xmm7

	 movd	  xmm7,.tex_ptr
	 shufps   xmm7,xmm7,0
	 movaps   [.tex_ptrI],xmm7

	 movd	  xmm1,.z1
	 shufps   xmm1,xmm1,0
	 movd	  xmm0,.dz
	 shufps   xmm0,xmm0,0
	 movaps   xmm7,xmm0		   ; xmm7 -> dz
	 cvtdq2ps xmm0,xmm0
	 mulps	  xmm0,[muler0123]
	 cvtps2dq xmm0,xmm0
	 paddd	  xmm1,xmm0
	 movaps   [.czI],xmm1	       ; xmm1 -> cz
	 pslld	  xmm7,2
	 movaps   [.dzI],xmm7

;         movaps   xmm7,[.cyI]
;         movaps   xmm6,[.cxI]
;         movaps   xmm5,[tex_cutter]
;         movaps   xmm4,[.tex_ptrI]

     .tl_loop:

	 movups     xmm0,[esi]
	 pcmpgtd    xmm0,xmm1	      ; packed compare Z buffer
	 pmovmskb   eax,xmm0	      ; check cause if all bits are zeroed
	 or	    eax,eax	      ; xmm0 -> mask
	 jz	    @f

	 movaps     xmm2,[.cyI]  ; xmm7
	 psrad	    xmm2,ROUND
	 pslld	    xmm2,TEX_SHIFT
	 movaps     xmm3,[.cxI]  ;xmm6
	 psrad	    xmm3,ROUND
	 paddd	    xmm2,xmm3
	 pand	    xmm2,[tex_cutter]	; xmm5
	 pslld	    xmm2,2
	 paddd	    xmm2,[.tex_ptrI]	; xmm4

if (Ext=SSE2)|(Ext=SSE3)
	 xchg	    esi,edi
	 maskmovdqu xmm1,xmm0	      ; actulize z buffer
	 xchg	    esi,edi
	 movd	    eax,xmm2

	 mov	    edx,[eax]
	 mov	    [.colI],edx
if 0
;;;;;;;-filtering
	 mov	     edx,[.cxI]        ; bilinear filtering
	 and	     edx,0x000000ff

   ;      neg         dl
   ;      add         dl,128
	 shl	     edx,7
      ;   mov        [.u1],dx
      ;   pshufw      mm5,[.u1],0
	 movd	     mm5,edx
	 punpcklwd mm5,mm5
	 punpckldq mm5,mm5

	 movd	     mm0,[eax]
	 movd	     mm3,[eax+4]
	 add	     eax,TEX_Y*4
	 movd	     mm2,[eax]
	 movd	     mm4,[eax+4]
	 pxor	     mm1,mm1
	 punpcklbw   mm0,mm1
	 punpcklbw   mm2,mm1

	 psubw	     mm2,mm0
	 pmulhuw     mm2,mm5
	 paddw	     mm0,mm2

	 punpcklbw   mm3,mm1
	 punpcklbw   mm4,mm1
	 psubw	     mm4,mm3
	 pmulhuw     mm4,mm5
	 paddw	     mm3,mm4

	 mov	     edx,[.cyI]
	 and	     edx,0x000000ff
   ;      neg         dl
   ;      add         dl,128
	 shl	     edx,7
      ;   mov        [.u1],dx
      ;   pshufw      mm5,[.u1],0
	 movd	     mm5,edx
	 punpcklwd   mm5,mm5
	 punpckldq   mm5,mm5

;         and         dx,0x00ff
   ;      neg         dl
   ;      add         dl,128
;         shl         dx,8
;         mov         [.v1],dx
;         pshufw      mm6,[.v1],0
	 psubw	     mm3,mm0
	 pmulhuw     mm3,mm6
	 paddw	     mm0,mm3
	 packuswb    mm0,mm0 ;   mm0 - > color
	 movd	     [.colI],mm0
 ;-----end of filtering
end if

	 psrldq     xmm2,4
	 movd	    eax,xmm2
	 mov	    edx,[eax]
	 mov	    [.colII],edx
if 0
;;;;;;;;;;;filtering
	 mov	     edx,[.cxII]	; bilinear filtering
;         and         dx,0x00ff
;         neg         dl
      ;   add         dl,128
;         shl         dx,8
;         mov         [.u1],dx
;         pshufw      mm5,[.u1],0
	 and	     edx,0x000000ff
	 shl	     edx,7
	 movd	     mm5,edx
	 punpcklwd   mm5,mm5
	 punpckldq   mm5,mm5

	 movd	     mm0,[eax]
	 movd	     mm3,[eax+4]
	 add	     eax,TEX_Y*4
	 movd	     mm2,[eax]
	 movd	     mm4,[eax+4]
	 pxor	     mm1,mm1
	 punpcklbw   mm0,mm1
	 punpcklbw   mm2,mm1

	 psubw	     mm2,mm0
	 pmulhuw     mm2,mm5
	 paddw	     mm0,mm2

	 punpcklbw   mm3,mm1
	 punpcklbw   mm4,mm1
	 psubw	     mm4,mm3
	 pmulhuw     mm4,mm5
	 paddw	     mm3,mm4

	 mov	     edx,[.cyII]
;         and         dx,0x00ff
	; neg         dl
	; add         dl,128
;         shl         dx,8
;         mov         [.v1],dx
;         pshufw      mm6,[.v1],0
	 and	     edx,0x000000ff
	 shl	     edx,7
	 movd	     mm5,edx
	 punpcklwd   mm5,mm5
	 punpckldq   mm5,mm5

	 psubw	     mm3,mm0
	 pmulhuw     mm3,mm6
	 paddw	     mm0,mm3
	 packuswb    mm0,mm0	     ;   mm0 - > color
	 movd	     [.colII],mm0
;;;;;end of filter
end if

	 psrldq     xmm2,4
	 movd	    eax,xmm2

	 mov	    edx,[eax]
	 mov	    [.colIII],edx
if 0
;;;;;filter
	 mov	     edx,[.cxIII]	 ; bilinear filtering
;         and         dx,0x00ff
       ;  neg         dl
       ;  add         dl,128
;         shl         dx,8
;         mov         [.u1],dx
;         pshufw      mm5,[.u1],0
	 and	     edx,0x000000ff
	 shl	     edx,7
	 movd	     mm5,edx
	 punpcklwd   mm5,mm5
	 punpckldq   mm5,mm5
	 movd	     mm0,[eax]
	 movd	     mm3,[eax+4]
	 add	     eax,TEX_Y*4
	 movd	     mm2,[eax]
	 movd	     mm4,[eax+4]
	 pxor	     mm1,mm1
	 punpcklbw   mm0,mm1
	 punpcklbw   mm2,mm1

	 psubw	     mm2,mm0
	 pmulhuw     mm2,mm5
	 paddw	     mm0,mm2

	 punpcklbw   mm3,mm1
	 punpcklbw   mm4,mm1
	 psubw	     mm4,mm3
	 pmulhuw     mm4,mm5
	 paddw	     mm3,mm4

	 mov	     edx,[.cyIII]
;         and         dx,0x00ff
       ;  neg         dl
       ;  add         dl,128
;         shl         dx,8
;         mov         [.v1],dx
;         pshufw      mm6,[.v1],0
	 and	     edx,0x000000ff
	 shl	     edx,7
	 movd	     mm5,edx
	 punpcklwd   mm5,mm5
	 punpckldq   mm5,mm5

	 psubw	     mm3,mm0
	 pmulhuw     mm3,mm6
	 paddw	     mm0,mm3
	 packuswb    mm0,mm0	     ;   mm0 - > color
	 movd	    [.colIII],mm0
;;;;;;;;;end of filter
end if

	 psrldq     xmm2,4
	 movd	    eax,xmm2
	 mov	    edx,[eax]
	 mov	    [.colIV],edx
if 0
;filteing
	 mov	     edx,[.cxIV]	; bilinear filtering
;         and         dx,0x00ff
      ;   neg         dl
      ;   add         dl,128
;         shl         dx,8
;         mov         [.u1],dx
;         pshufw      mm5,[.u1],0
	 and	     edx,0x000000ff
	 shl	     edx,7
	 movd	     mm5,edx
	 punpcklwd   mm5,mm5
	 punpckldq   mm5,mm5

	 movd	     mm0,[eax]
	 movd	     mm3,[eax+4]
	 add	     eax,TEX_Y*4
	 movd	     mm2,[eax]
	 movd	     mm4,[eax+4]
	 pxor	     mm1,mm1
	 punpcklbw   mm0,mm1
	 punpcklbw   mm2,mm1

	 psubw	     mm2,mm0
	 pmulhuw     mm2,mm5
	 paddw	     mm0,mm2

	 punpcklbw   mm3,mm1
	 punpcklbw   mm4,mm1
	 psubw	     mm4,mm3
	 pmulhuw     mm4,mm5
	 paddw	     mm3,mm4

	 mov	     edx,[.cyIV]
;         and         dx,0x00ff
       ;  neg         dl
;         shl         dx,8
;         mov         [.v1],dx
;         pshufw      mm6,[.v1],0
	 and	     edx,0x000000ff
	 shl	     edx,7
	 movd	     mm5,edx
	 punpcklwd   mm5,mm5
	 punpckldq   mm5,mm5

	 psubw	     mm3,mm0
	 pmulhuw     mm3,mm6
	 paddw	     mm0,mm3
	 packuswb    mm0,mm0	     ;   mm0 - > color
	 movd	     [.colIV],mm0
;;;;;;;;;;end of filter
end if
	 movaps     xmm2,[.colI]
	 maskmovdqu xmm2,xmm0	      ; put conditionally 4 pixels
else
       movaps	  [.cyI],xmm7
       movd	  eax,xmm2
       pinsrd	  xmm7,[eax],0 ; insert double word into first position
       psrldq	  xmm2,4
       movd	  eax,xmm2
       pinsrd	  xmm7,[eax],1 ; insert double word into second position
       psrldq	  xmm2,4
       movd	  eax,xmm2
       pinsrd	  xmm7,[eax],2 ; insert double word into 3th position
       psrldq	  xmm2,4
       movd	  eax,xmm2
       pinsrd	  xmm7,[eax],3 ; insert double word into 4th position
       movaps	  [.cxI],xmm6
       movups	  xmm6,[edi]
       pblendvb   xmm6,xmm7	    ; draw color
       movaps	  xmm7,[.cyI]
       movups	  [edi],xmm6
       movups	  xmm6,[esi]
       pblendvb   xmm6,xmm1	    ; actualize z buffer
       movups	  [esi],xmm6
       movaps	  xmm6,[.cxI]
end if
    @@:
	 add	    edi,16
	 add	    esi,16
	 movaps   xmm7,[.cyI]
	 movaps   xmm6,[.cxI]
	 paddd	  xmm6,[.tex_dxI]
	 paddd	  xmm7,[.tex_dyI]
	 paddd	  xmm1,[.dzI]
	 movaps   [.cyI],xmm7
	 movaps   [.cxI],xmm6

	 dec	    ecx
	 jnz	    .tl_loop

   .draw_last:
	movd	.cz,xmm1
	movd	.c_tex_x,xmm6
	movd	.c_tex_y,xmm7

   .draw_last1:
	movzx	ecx,word[.rest]
	or	ecx,ecx
	jz   .tl_quit
	mov	edx,.cz
    .ddraw:
	cmp	[esi],edx
	jle	@f
	mov	[esi],edx  ;renew z buffer
	mov	ebx,.c_tex_x
	sar	ebx,ROUND
	mov	eax,.c_tex_y
	sar	eax,ROUND
	shl	eax,TEX_SHIFT
	add	eax,ebx
	shl	eax,2
	add	eax,.tex_ptr
	mov	eax,[eax]
 if 0
	mov	ebx,.c_tex_x	   ; I tried bilinear filtering
	and	bx,0x00ff
	shl	bx,8
	mov	[.u],bx
	pshufw	mm5,[.u],0

	movd	    mm0,[eax]
	movd	    mm3,[eax+4]
	add	    eax,TEX_Y*4
	movd	    mm2,[eax]
	movd	    mm4,[eax+4]
	pxor	    mm1,mm1
	punpcklbw   mm0,mm1
	punpcklbw   mm2,mm1

	psubw	    mm2,mm0
	pmulhuw     mm2,mm5
	paddw	    mm0,mm2

	punpcklbw   mm3,mm1
	punpcklbw   mm4,mm1
	psubw	    mm4,mm3
	pmulhuw     mm4,mm5
	paddw	    mm3,mm4

	mov	    ebx,.c_tex_y
	and	    bx,0x00ff
	shl	    bx,8
	mov	    [.v],bx
	pshufw	    mm6,[.v],0
	psubw	    mm3,mm0
	pmulhuw     mm3,mm6
	paddw	    mm0,mm3
	packuswb    mm0,mm0	    ;   mm0 - > color
end if
	mov    [edi],eax ;mm0

      @@:
	add	esi,4
	add	edi,4
	add	edx,.dz
	mov	eax,.tex_dx
	add	.c_tex_x,eax
	mov	eax,.tex_dy
	add	.c_tex_y,eax
	dec	ecx
	jnz	.ddraw

   .tl_quit:

	mov	esp,ebp

ret 30+8

